@@ -6,6 +6,75 @@
  * of the GNU General Public License v2 or (at your option) any later version.
  */
 
+/*
+ * . application in the kernel or userspace asks to join
+ *   a lockspace
+ *
+ * . dlm_new_lockspace() in the kernel sets up basic
+ *   lockspace structures, then sends a uevent to
+ *   dlm_controld in userspace (KOBJ_ONLINE), asking
+ *   dlm_controld to join the corosync group ("cpg")
+ *   for the lockspace.  dlm_new_lockspace() blocks
+ *   waiting for a reply from dlm_controld (the reply is
+ *   a write to the "event_done" sysfs file).
+ *   When the successful reply comes back to dlm-kernel
+ *   from dlm_controld, dlm-kernel knows it is now a member
+ *   of the lockspace membership (represented in userspace
+ *   by the corosync cpg), and can do locking with the
+ *   other members.  Before sending event_done to the kernel,
+ *   dlm_controld tells dlm-kernel who the other lockspace
+ *   members are via configfs settings.
+ *
+ * . When dlm_controld gets a request from dlm-kernel to
+ *   join a lockspace, it runs dlm_join_lockspace() which
+ *   calls cpg_join() to join the corosync group representing
+ *   the members of the lockspace.  dlm_controld will get
+ *   callbacks from corosync when membership of this cpg 
+ *   changes (joins/leaves/failures).  After calling
+ *   cpg_join(), dlm_controld waits for the first
+ *   corosync membership callback indicating it is now
+ *   a member of the cpg.  The callback function for
+ *   cpg membership changes is confchg_cb().  Corosync
+ *   guarantees that all members of the cpg see the
+ *   same sequence of confchg callbacks (e.g. if a number
+ *   of nodes are joining/leaving/failing at once).
+ *   When the first confchg arrives after cpg_join(),
+ *   dlm_controld sets up the current members for dlm-kernel
+ *   via configfs, then writes to event_done in sysfs to start
+ *   dlm-kernel running.
+ *
+ * . When a cpg member joins/leaves/fails, dlm_controld
+ *   on all current members gets a confchg callback
+ *   showing the new members.  dlm_controld then stops
+ *   dlm-kernel locking activity for that lockspace by
+ *   writing 0 to the "control" sysfs file.
+ *   dlm_controld then sends/recvs cpg messages to all
+ *   other cpg members to act as barrier to ensure all
+ *   members have stopped locking activity in the kernel
+ *   (apply_changes()).  When all have done this,
+ *   dlm_controld on all the members then sets up the
+ *   new members in the kernel (via configfs) and tells
+ *   dlm-kernel to start the lockspace again (start_kernel()).
+ *
+ * . When dlm-kernel is started after being stopped, it does
+ *   lockspace recovery based on changes to the membership.
+ *   When recovery is done, normal locking activity resumes.
+ *
+ * Replacing dlm_controld is a matter doing the following
+ * steps by either manually setting up sysfs and configfs,
+ * or having a new daemon to do it:
+ *
+ * - decide who the lockspace members are
+ * - stop dlm-kernel before changing lockspace members (write to sysfs)
+ * - wait for previous step on all before making changes
+ * - tell dlm-kernel member nodeids/IPs in configfs (write to configfs)
+ * - start dlm-kernel (write to sysfs)
+ *
+ * echo 0/1 into /sys/kernel/dlm/foo/control and /sys/kernel/dlm/foo/event_done
+ * echo/mkdir/write values into /sys/kernel/config/dlm/cluster/comms/ and
+ * /sys/kernel/config/dlm/cluster/spaces/foo/
+ */
+
 #include "dlm_daemon.h"
 
 #define log_limit(ls, fmt, args...) ({        \
@@ -381,21 +450,24 @@ static int check_ringid_done(struct lockspace *ls)
 	   but that's probably not guaranteed.) */
 
 	if (ls->cpg_ringid_wait) {
-		log_group(ls, "check_ringid wait cluster %u cpg %u:%llu",
-			  cluster_ringid_seq, ls->cpg_ringid.nodeid,
+		log_group(ls, "check_ringid wait cluster %u cpg %llu:%llu",
+			  (unsigned long long)cluster_ringid_seq,
+			  ls->cpg_ringid.nodeid,
 			  (unsigned long long)ls->cpg_ringid.seq);
 		return 0;
 	}
 
-	if (cluster_ringid_seq != (uint32_t)ls->cpg_ringid.seq) {
-		log_group(ls, "check_ringid cluster %u cpg %u:%llu",
-			  cluster_ringid_seq, ls->cpg_ringid.nodeid,
+	if (cluster_ringid_seq != ls->cpg_ringid.seq) {
+		log_group(ls, "check_ringid cluster %llu cpg %u:%llu",
+			  (unsigned long long)cluster_ringid_seq,
+			  ls->cpg_ringid.nodeid,
 			  (unsigned long long)ls->cpg_ringid.seq);
 		return 0;
 	}
 
-	log_limit(ls, "check_ringid done cluster %u cpg %u:%llu",
-		  cluster_ringid_seq, ls->cpg_ringid.nodeid,
+	log_limit(ls, "check_ringid done cluster %llu cpg %u:%llu",
+		  (unsigned long long)cluster_ringid_seq,
+		  ls->cpg_ringid.nodeid,
 		  (unsigned long long)ls->cpg_ringid.seq);
 
 	return 1;
@@ -1664,7 +1736,7 @@ static void process_cpg_lockspace(int ci)
 	}
 
 	error = cpg_dispatch(ls->cpg_handle, CS_DISPATCH_ALL);
-	if (error != CS_OK) {
+	if (error != CS_OK && error != CS_ERR_BAD_HANDLE) {
 		log_error("cpg_dispatch error %d", error);
 		return;
 	}
@@ -1795,7 +1867,8 @@ int set_lockspace_info(struct lockspace *ls, struct dlmc_lockspace *lockspace)
 {
 	struct change *cg, *last = NULL;
 
-	strncpy(lockspace->name, ls->name, DLM_LOCKSPACE_LEN);
+	strncpy(lockspace->name, ls->name, DLM_LOCKSPACE_LEN + 1);
+	lockspace->name[DLM_LOCKSPACE_LEN] = '\0';
 	lockspace->global_id = ls->global_id;
 
 	if (ls->joining)
