@@ -79,6 +79,7 @@ struct node_daemon {
 	int fence_actor_done; /* for status/debug */
 	int fence_actor_last; /* for status/debug */
 	int fence_actors[MAX_NODES];
+	int fence_actors_orig[MAX_NODES];
 
 	struct protocol proto;
 	struct fence_config fence_config;
@@ -107,7 +108,6 @@ static int daemon_remove_count;
 static int daemon_ringid_wait;
 static struct cpg_ring_id daemon_ringid;
 static int daemon_fence_pid;
-static uint64_t daemon_last_join_monotime;
 static uint32_t last_join_seq;
 static uint32_t send_fipu_seq;
 static int wait_clear_fipu;
@@ -119,6 +119,7 @@ static int zombie_count;
 
 static int fence_result_pid;
 static unsigned int fence_result_try;
+static int stateful_merge_wait; /* cluster is stuck in waiting for manual intervention */
 
 static void send_fence_result(int nodeid, int result, uint32_t flags, uint64_t walltime);
 static void send_fence_clear(int nodeid, int result, uint32_t flags, uint64_t walltime);
@@ -173,7 +174,8 @@ void log_config(const struct cpg_name *group_name,
 
 	log_debug("%s conf %zu %zu %zu memb%s join%s left%s", group_name->value,
 		  member_list_entries, joined_list_entries, left_list_entries,
-		  m_buf, j_buf, l_buf);
+		  strlen(m_buf) ? m_buf : " 0", strlen(j_buf) ? j_buf : " 0",
+		  strlen(l_buf) ? l_buf : " 0");
 }
 
 void log_ringid(const char *name,
@@ -314,6 +316,24 @@ void dlm_send_message(struct lockspace *ls, char *buf, int len)
 	_send_message(ls->cpg_handle, buf, len, type);
 }
 
+int dlm_send_message_daemon(char *buf, int len)
+{
+	struct dlm_header *hd = (struct dlm_header *) buf;
+	int type = hd->type;
+
+	hd->version[0]  = cpu_to_le16(our_protocol.daemon_run[0]);
+	hd->version[1]  = cpu_to_le16(our_protocol.daemon_run[1]);
+	hd->version[2]  = cpu_to_le16(our_protocol.daemon_run[2]);
+	hd->type	= cpu_to_le16(hd->type);
+	hd->nodeid      = cpu_to_le32(our_nodeid);
+	hd->to_nodeid   = cpu_to_le32(hd->to_nodeid);
+	hd->flags       = cpu_to_le32(hd->flags);
+	hd->msgdata     = cpu_to_le32(hd->msgdata);
+	hd->msgdata2    = cpu_to_le32(hd->msgdata2);
+
+	return _send_message(cpg_handle_daemon, buf, len, type);
+}
+
 void dlm_header_in(struct dlm_header *hd)
 {
 	hd->version[0]  = le16_to_cpu(hd->version[0]);
@@ -328,6 +348,50 @@ void dlm_header_in(struct dlm_header *hd)
 	hd->msgdata2    = le32_to_cpu(hd->msgdata2);
 }
 
+static void run_info_out(struct run_info *info)
+{
+	info->dest_nodeid  = cpu_to_le32(info->dest_nodeid);
+	info->start_nodeid = cpu_to_le32(info->start_nodeid);
+	info->local_pid    = cpu_to_le32(info->local_pid);
+	info->local_result = cpu_to_le32(info->local_result);
+	info->need_replies = cpu_to_le32(info->need_replies);
+	info->reply_count  = cpu_to_le32(info->reply_count);
+	info->fail_count   = cpu_to_le32(info->fail_count);
+	info->flags        = cpu_to_le32(info->flags);
+}
+
+static void run_info_in(struct run_info *info)
+{
+	info->dest_nodeid  = le32_to_cpu(info->dest_nodeid);
+	info->start_nodeid = le32_to_cpu(info->start_nodeid);
+	info->local_pid    = le32_to_cpu(info->local_pid);
+	info->local_result = le32_to_cpu(info->local_result);
+	info->need_replies = le32_to_cpu(info->need_replies);
+	info->reply_count  = le32_to_cpu(info->reply_count);
+	info->fail_count   = le32_to_cpu(info->fail_count);
+	info->flags        = le32_to_cpu(info->flags);
+}
+
+static void run_request_out(struct run_request *req)
+{
+	run_info_out(&req->info);
+}
+
+static void run_request_in(struct run_request *req)
+{
+	run_info_in(&req->info);
+}
+
+static void run_reply_out(struct run_reply *rep)
+{
+	run_info_out(&rep->info);
+}
+
+static void run_reply_in(struct run_reply *rep)
+{
+	run_info_in(&rep->info);
+}
+
 int dlm_header_validate(struct dlm_header *hd, int nodeid)
 {
 	if (hd->version[0] != our_protocol.daemon_run[0] ||
@@ -537,6 +601,7 @@ static int set_fence_actors(struct node_daemon *node, int all_memb)
 	int i, nodeid, count = 0, low = 0;
 
 	memset(node->fence_actors, 0, sizeof(node->fence_actors));
+	memset(node->fence_actors_orig, 0, sizeof(node->fence_actors_orig));
 
 	for (i = 0; i < daemon_member_count; i++) {
 		nodeid = daemon_member[i].nodeid;
@@ -550,6 +615,9 @@ static int set_fence_actors(struct node_daemon *node, int all_memb)
 			low = nodeid;
 	}
 
+	/* keep a copy of the original set so they can be retried if all fail */
+	memcpy(node->fence_actors_orig, node->fence_actors, sizeof(node->fence_actors));
+
 	log_debug("set_fence_actors for %d low %d count %d",
 		  node->nodeid, low, count);
 	return low;
@@ -592,6 +660,7 @@ static int get_fence_actor(struct node_daemon *node)
 static void clear_fence_actor(int nodeid, int actor)
 {
 	struct node_daemon *node;
+	int remaining = 0;
 	int i;
 
 	node = get_node_daemon(nodeid);
@@ -599,10 +668,15 @@ static void clear_fence_actor(int nodeid, int actor)
 		return;
 
 	for (i = 0; i < MAX_NODES; i++) {
-		if (node->fence_actors[i] == actor) {
+		if (node->fence_actors[i] == actor)
 			node->fence_actors[i] = 0;
-			return;
-		}
+		else if (node->fence_actors[i])
+			remaining++;
+	}
+
+	if (!remaining && opt(repeat_failed_fencing_ind)) {
+		log_debug("clear_fence_actor %d restoring original actors to retry", actor);
+		memcpy(node->fence_actors, node->fence_actors_orig, sizeof(node->fence_actors));
 	}
 }
 
@@ -766,6 +840,9 @@ static void daemon_fence_work(void)
 	int retry = 0;
 	uint32_t flags;
 
+	if (!daemon_fence_allow)
+		return;
+
 	if (daemon_ringid_wait) {
 		/* We've seen a nodedown confchg callback, but not the
 		   corresponding ringid callback. */
@@ -845,10 +922,14 @@ static void daemon_fence_work(void)
 
 		if ((clean_count >= merge_count) && !part_count && (low == our_nodeid))
 			kick_stateful_merge_members();
+		if ((clean_count < merge_count) && !part_count)
+			stateful_merge_wait = 1;
 
 		retry = 1;
 		goto out;
 	}
+	if (stateful_merge_wait)
+		stateful_merge_wait = 0;
 
 	/*
 	 * startup fencing
@@ -865,10 +946,15 @@ static void daemon_fence_work(void)
 		if (!opt(enable_startup_fencing_ind))
 			continue;
 
-		if (monotime() - daemon_last_join_monotime < opt(post_join_delay_ind)) {
+		if (!fence_delay_begin) {
+			log_debug("fence startup %d wait for initial delay", node->nodeid);
+			continue;
+		}
+
+		if (monotime() - fence_delay_begin < opt(post_join_delay_ind)) {
 			log_debug("fence startup %d delay %d from %llu",
 				  node->nodeid, opt(post_join_delay_ind),
-				  (unsigned long long)daemon_last_join_monotime);
+				  (unsigned long long)fence_delay_begin);
 			retry = 1;
 			continue;
 		}
@@ -951,10 +1037,10 @@ static void daemon_fence_work(void)
 		   time between it joining the cluster (giving cluster quorum)
 		   and joining the daemon cpg, which allows it to bypass fencing */
 
-		if (monotime() - cluster_last_join_monotime < opt(post_join_delay_ind)) {
+		if (monotime() - fence_delay_begin < opt(post_join_delay_ind)) {
 			log_debug("fence request %d delay %d from %llu",
 				  node->nodeid, opt(post_join_delay_ind),
-				  (unsigned long long)cluster_last_join_monotime);
+				  (unsigned long long)fence_delay_begin);
 			node->delay_fencing = 1;
 			retry = 1;
 			continue;
@@ -1806,6 +1892,7 @@ int set_protocol(void)
 {
 	struct protocol proto;
 	struct pollfd pollfd;
+	cs_error_t error;
 	int sent_proposal = 0;
 	int rv;
 
@@ -1855,8 +1942,17 @@ int set_protocol(void)
 			return -1;
 		}
 
-		if (pollfd.revents & POLLIN)
-			process_cpg_daemon(0);
+		if (pollfd.revents & POLLIN) {
+			/*
+			 * don't use process_cpg_daemon() because we only want to
+			 * dispatch one thing at a time because we only want to
+			 * handling protocol related things here.
+			 */
+
+			error = cpg_dispatch(cpg_handle_daemon, CS_DISPATCH_ONE);
+			if (error != CS_OK && error != CS_ERR_BAD_HANDLE)
+				log_error("daemon cpg_dispatch one error %d", error);
+		}
 		if (pollfd.revents & (POLLERR | POLLHUP | POLLNVAL)) {
 			log_error("set_protocol poll revents %u",
 				  pollfd.revents);
@@ -1922,6 +2018,15 @@ static void deliver_cb_daemon(cpg_handle_t handle,
 	hd = (struct dlm_header *)data;
 	dlm_header_in(hd);
 
+	if (!daemon_fence_allow && hd->type != DLM_MSG_PROTOCOL) {
+		/* don't think this will happen; if it does we may
+		   need to verify that it's correct to ignore these
+		   messages instead of saving them to process after
+		   allow is set */
+		log_debug("deliver_cb_daemon ignore non proto msg %d", hd->type);
+		return;
+	}
+
 	switch (hd->type) {
 	case DLM_MSG_PROTOCOL:
 		receive_protocol(hd, len);
@@ -1932,6 +2037,12 @@ static void deliver_cb_daemon(cpg_handle_t handle,
 	case DLM_MSG_FENCE_CLEAR:
 		receive_fence_clear(hd, len);
 		break;
+	case DLM_MSG_RUN_REQUEST:
+		receive_run_request(hd, len);
+		break;
+	case DLM_MSG_RUN_REPLY:
+		receive_run_reply(hd, len);
+		break;
 	default:
 		log_error("deliver_cb_daemon unknown msg type %d", hd->type);
 	}
@@ -1939,6 +2050,204 @@ static void deliver_cb_daemon(cpg_handle_t handle,
 	daemon_fence_work();
 }
 
+int receive_run_reply(struct dlm_header *hd, int len)
+{
+	struct run_reply *rep = (struct run_reply *)hd;
+	struct run *run;
+	int i;
+
+	run_reply_in(rep);
+
+	log_debug("receive_run_reply %s from %d result %d",
+		  rep->uuid, hd->nodeid, rep->info.local_result);
+
+	if (!opt(enable_helper_ind)) {
+		log_debug("receive_run_reply %s helper not enabled", rep->uuid);
+		return 0;
+	}
+
+	run = find_run(rep->uuid);
+	if (!run) {
+		log_debug("receive_run_reply no uuid %s", rep->uuid);
+		return 0;
+	}
+
+	/*
+	 * Only the starting node keeps track of results.
+	 */
+	if (run->info.start_nodeid != our_nodeid)
+		return 0;
+
+	if (len != sizeof(struct run_reply)) {
+		log_debug("receive_run_reply %s bad len %s expect %d",
+			  rep->uuid, len, sizeof(struct run_reply));
+		run->info.reply_count++;
+		run->info.need_replies--;
+		return 0;
+	}
+
+	for (i = 0; i < run->node_count; i++) {
+		if (run->node_results[i].nodeid != hd->nodeid)
+			continue;
+
+		/* shouldn't happen? */
+		if (run->node_results[i].replied)
+			break;
+
+		run->node_results[i].result = rep->info.local_result;
+		run->node_results[i].replied = 1;
+
+		if (rep->info.local_result)
+			run->info.fail_count++;
+
+		run->info.reply_count++;
+		run->info.need_replies--;
+
+		/*
+		log_debug("run reply_count % need_replies %d fail_count %d",
+			  run->info.reply_count, run->info.need_replies, run->info.fail_count);
+		*/
+		break;
+	}
+
+	return 0;
+}
+
+int receive_run_request(struct dlm_header *hd, int len)
+{
+	struct run_request *req = (struct run_request *)hd;
+	struct run *run = NULL;
+
+	run_request_in(req);
+
+	log_debug("receive_run_request %s from %d", req->uuid, hd->nodeid);
+
+	if (len != sizeof(struct run_request)) {
+		log_debug("receive_run_request %s bad len %d", req->uuid, len);
+		/* todo: send reply with failed */
+		return 0;
+	}
+
+	if (req->info.dest_nodeid && (req->info.dest_nodeid != our_nodeid))
+		return 0;
+
+	if (req->info.start_nodeid == our_nodeid) {
+		if (!(req->info.flags & DLMC_FLAG_RUN_START_NODE_RECV)) {
+			log_debug("receive_run_request ignore self");
+			return 0;
+		}
+
+		if (!opt(enable_helper_ind)) {
+			log_debug("receive_run_request %s helper not enabled", req->uuid);
+			return 0;
+		}
+
+		run = find_run(req->uuid);
+		if (!run) {
+			log_debug("receive_run_request from self no uuid %s", req->uuid);
+			return 0;
+		}
+
+		log_debug("receive_run_request %s to helper", req->uuid);
+
+		send_helper_run_request(req);
+		return 0;
+	}
+
+	if (!opt(enable_helper_ind) && run) {
+		log_debug("receive_run_request %s helper not enabled", req->uuid);
+		run->info.reply_count++;
+		run->info.need_replies--;
+		/* todo: send reply with failed */
+		return 0;
+	}
+
+	if (!(run = malloc(sizeof(struct run)))) {
+		log_error("receive_run_request %s no mem", req->uuid);
+		/* todo: send reply with failed */
+		return 0;
+	}
+
+	memset(run, 0, sizeof(struct run));
+
+	memcpy(run->uuid, req->uuid, RUN_UUID_LEN);
+	memcpy(run->command, req->command, RUN_COMMAND_LEN);
+	run->info.start_nodeid = req->info.start_nodeid;
+	run->info.dest_nodeid = req->info.dest_nodeid;
+	run->info.flags = req->info.flags;
+
+	list_add(&run->list, &run_ops);
+
+	log_error("run request %s %.128s", run->uuid, run->command);
+
+	log_debug("receive_run_request %s to helper", req->uuid);
+
+	send_helper_run_request(req);
+	/* todo: if no helper, send reply with failed */
+
+	return 0;
+}
+
+int send_run_request(struct run *run, struct run_request *req)
+{
+	struct node_daemon *node;
+	int i = 0;
+	int rv;
+
+	list_for_each_entry(node, &daemon_nodes, list) {
+		if (!node->daemon_member)
+			continue;
+
+		/*
+		 * When this starting node does not run the command,
+		 * there is no reply for our nodeid.
+		 */
+		if ((node->nodeid == our_nodeid) &&
+		    (run->info.flags & DLMC_FLAG_RUN_START_NODE_NONE))
+				continue;
+
+		/*
+		 * The command is only run on one specific node, and
+		 * only a reply from that node is needed.
+		 */
+		if (run->info.dest_nodeid && (node->nodeid != run->info.dest_nodeid))
+			continue;
+
+		run->node_count++;
+		run->node_results[i].nodeid = node->nodeid;
+		i++;
+	}
+
+	run->info.need_replies = run->node_count;
+
+	log_debug("send_run_request %s for %d nodes", req->uuid, run->node_count);
+
+	run_request_out(req);
+
+	rv = dlm_send_message_daemon((char *)req, sizeof(struct run_request));
+
+	return rv;
+}
+
+int send_run_reply(struct run *run, struct run_reply *rep)
+{
+	int rv;
+
+	log_debug("send_run_reply %s result %d", rep->uuid, rep->info.local_result);
+
+	run_reply_out(rep);
+
+	rv = dlm_send_message_daemon((char *)rep, sizeof(struct run_reply));
+
+	/*
+	 * If we are not the starting node, clear the run operation.
+	 */
+	if (rep->info.start_nodeid != our_nodeid)
+		clear_run(run);
+
+	return rv;
+}
+
 static void confchg_cb_daemon(cpg_handle_t handle,
 			      const struct cpg_name *group_name,
 			      const struct cpg_address *member_list,
@@ -2015,7 +2324,7 @@ static void confchg_cb_daemon(cpg_handle_t handle,
 			node->daemon_member = 1;
 			node->daemon_add_time = now;
 
-			daemon_last_join_monotime = now;
+			fence_delay_begin = now;
 			last_join_seq++;
 
 			/* a joining node shows prev members in joined list */
@@ -2027,7 +2336,7 @@ static void confchg_cb_daemon(cpg_handle_t handle,
 				   valid proto from it (is_clean_daemon_member) */
 				log_error("daemon joined %d needs fencing", node->nodeid);
 			} else {
-				log_debug("daemon joined %d");
+				log_debug("daemon joined %d", node->nodeid);
 			}
 		} else {
 			if (!node->daemon_member)
@@ -2155,7 +2464,7 @@ void process_cpg_daemon(int ci)
 	cs_error_t error;
 
 	error = cpg_dispatch(cpg_handle_daemon, CS_DISPATCH_ALL);
-	if (error != CS_OK)
+	if (error != CS_OK && error != CS_ERR_BAD_HANDLE)
 		log_error("daemon cpg_dispatch error %d", error);
 }
 
@@ -2356,7 +2665,8 @@ static int print_state_daemon(char *str)
 		 "fence_pid=%d "
 		 "fence_in_progress_unknown=%d "
 		 "zombie_count=%d "
-		 "monotime=%llu ",
+		 "monotime=%llu "
+		 "stateful_merge_wait=%d ",
 		 daemon_member_count,
 		 daemon_joined_count,
 		 daemon_remove_count,
@@ -2366,7 +2676,8 @@ static int print_state_daemon(char *str)
 		 daemon_fence_pid,
 		 fence_in_progress_unknown,
 		 zombie_count,
-		 (unsigned long long)monotime());
+		 (unsigned long long)monotime(),
+		 stateful_merge_wait);
 
 	return strlen(str) + 1;
 }
