@@ -41,6 +41,14 @@
 #define OP_FENCE_ACK			11
 #define OP_STATUS			12
 #define OP_DUMP_CONFIG			13
+#define OP_RUN				14
+#define OP_RUN_START			15
+#define OP_RUN_CHECK			16
+#define OP_RUN_CANCEL			17
+#define OP_RUN_LIST			18
+#define OP_DUMP_RUN			19
+#define OP_RELOAD_CONFIG	20
+#define OP_SET_CONFIG		21
 
 static char *prog_name;
 static char *lsname;
@@ -53,8 +61,12 @@ static int dump_mstcpy = 0;
 static mode_t create_mode = 0600;
 static int verbose;
 static int wide;
+static int wait_sec;
 static int summarize;
 
+char run_command[DLMC_RUN_COMMAND_LEN];
+char run_uuid[DLMC_RUN_UUID_LEN];
+
 #define MAX_LS 128
 #define MAX_NODES 128
 
@@ -186,9 +198,11 @@ static void print_usage(void)
 	printf("dlm_tool [command] [options] [name]\n");
 	printf("\n");
 	printf("Commands:\n");
-	printf("ls, status, dump, dump_config, fence_ack\n");
+	printf("ls, status, dump, fence_ack\n");
+	printf("dump_config, reload_config, set_config\n");
 	printf("log_plock, plocks\n");
 	printf("join, leave, lockdebug\n");
+	printf("run, run_start, run_check, run_cancel, run_list\n");
 	printf("\n");
 	printf("Options:\n");
 	printf("  -n               Show all node information in ls\n");
@@ -198,18 +212,23 @@ static void print_usage(void)
 	printf("  -s               Summary following lockdebug output (experimental)\n");
 	printf("  -v               Verbose lockdebug output\n");
 	printf("  -w               Wide lockdebug output\n");
+	printf("  -i <sec>         Wait for <sec>.\n");
 	printf("  -h               Print help, then exit\n");
 	printf("  -V               Print program version information, then exit\n");
 	printf("\n");
 }
 
-#define OPTION_STRING "MhVnm:e:f:vws"
+#define OPTION_STRING "MhVnm:e:f:vwsi:"
 
 static void decode_arguments(int argc, char **argv)
 {
 	int cont = 1;
 	int optchar;
 	int need_lsname;
+	int need_command = 0;
+	int need_uuid = 0;
+	int optional_lsname = 0;
+	int i;
 	char modebuf[8];
 
 	while (cont) {
@@ -250,6 +269,10 @@ static void decode_arguments(int argc, char **argv)
 			wide = 1;
 			break;
 
+		case 'i':
+			wait_sec = atoi(optarg);
+			break;
+
 		case 'h':
 			print_usage();
 			exit(EXIT_SUCCESS);
@@ -313,6 +336,7 @@ static void decode_arguments(int argc, char **argv)
 			operation = OP_LIST;
 			opt_ind = optind + 1;
 			need_lsname = 0;
+			optional_lsname = 1;
 			break;
 		} else if (!strncmp(argv[optind], "status", 6) &&
 			   (strlen(argv[optind]) == 6)) {
@@ -342,6 +366,19 @@ static void decode_arguments(int argc, char **argv)
 			opt_ind = optind + 1;
 			need_lsname = 0;
 			break;
+		} else if (!strncmp(argv[optind], "reload_config", 13) &&
+			   (strlen(argv[optind]) == 13)) {
+			operation = OP_RELOAD_CONFIG;
+			opt_ind = optind + 1;
+			need_lsname = 0;
+			break;
+		} else if (!strncmp(argv[optind], "set_config", 10) &&
+			   (strlen(argv[optind]) == 10)) {
+			operation = OP_SET_CONFIG;
+			opt_ind = optind + 1;
+			need_lsname = 0;
+			need_command = 1;
+			break;
 		} else if (!strncmp(argv[optind], "plocks", 6) &&
 			   (strlen(argv[optind]) == 6)) {
 			operation = OP_PLOCKS;
@@ -353,8 +390,39 @@ static void decode_arguments(int argc, char **argv)
 			opt_ind = optind + 1;
 			need_lsname = 0;
 			break;
+		} else if (!strcmp(argv[optind], "run")) {
+			operation = OP_RUN;
+			opt_ind = optind + 1;
+			need_lsname = 0;
+			need_command = 1;
+			break;
+		} else if (!strcmp(argv[optind], "run_start")) {
+			operation = OP_RUN_START;
+			opt_ind = optind + 1;
+			need_lsname = 0;
+			need_command = 1;
+			break;
+		} else if (!strcmp(argv[optind], "run_check")) {
+			operation = OP_RUN_CHECK;
+			opt_ind = optind + 1;
+			need_lsname = 0;
+			need_uuid = 1;
+			break;
+		} else if (!strcmp(argv[optind], "run_cancel")) {
+			operation = OP_RUN_CANCEL;
+			opt_ind = optind + 1;
+			need_lsname = 0;
+			need_uuid = 1;
+			break;
+		} else if (!strcmp(argv[optind], "run_list")) {
+			operation = OP_DUMP_RUN;
+			opt_ind = optind + 1;
+			need_lsname = 0;
+			need_uuid = 0;
+			break;
 		}
 
+
 		/*
 		 * debugfs
 		 */
@@ -378,11 +446,54 @@ static void decode_arguments(int argc, char **argv)
 		exit(EXIT_FAILURE);
 	}
 
-	if (optind < argc - 1)
-		lsname = argv[opt_ind];
-	else if (need_lsname) {
-		fprintf(stderr, "lockspace name required\n");
+	/*
+	 * The operation does not require an arg, but may
+	 * accept an optional arg, i.e. dlm_tool ls [name]
+	 */
+	if (!need_lsname && !need_uuid && !need_command) {
+		if (optind < argc - 1) {
+			if (optional_lsname)
+				lsname = argv[opt_ind];
+		}
+		return;
+	}
+
+	if (optind < argc - 1) {
+		if (need_lsname)
+			lsname = argv[opt_ind];
+		else if (need_uuid) {
+			strncpy(run_uuid, argv[opt_ind], DLMC_RUN_UUID_LEN - 1);
+			run_uuid[DLMC_RUN_UUID_LEN - 1] = '\0';
+		} else if (need_command)
+			goto copy_command;
+	} else if (need_lsname) {
+		if (operation == OP_FENCE_ACK)
+			fprintf(stderr, "nodeid required\n");
+		else
+			fprintf(stderr, "lockspace name required\n");
 		exit(EXIT_FAILURE);
+	} else if (need_command) {
+		fprintf(stderr, "command required\n");
+		exit(EXIT_FAILURE);
+	} else if (need_uuid) {
+		fprintf(stderr, "uuid required\n");
+		exit(EXIT_FAILURE);
+	} else {
+		fprintf(stderr, "missing arg\n");
+		exit(EXIT_FAILURE);
+	}
+
+ copy_command:
+	for (i = opt_ind; i < argc; i++) {
+		if (strlen(run_command) + strlen(argv[i]) + 2 > DLMC_RUN_COMMAND_LEN) {
+			fprintf(stderr, "command too long (max %d)\n", DLMC_RUN_COMMAND_LEN);
+			exit(EXIT_FAILURE);
+		}
+
+		if (strlen(argv[i])) {
+			strcat(run_command, argv[i]);
+			strcat(run_command, " ");
+		}
 	}
 }
 
@@ -659,10 +770,10 @@ static char *pr_verbose(struct lkb *lkb)
 
 	memset(buf, 0, sizeof(buf));
 
-	sprintf(buf, "time %016llu flags %08x %08x bast %d %llu",
+	sprintf(buf, "time %016llu flags %08x %08x bast %d %llu pid %d",
 		(unsigned long long)lkb->timestamp,
 		lkb->exflags, lkb->flags, lkb->highbast,
-		(unsigned long long)lkb->time_bast);
+		(unsigned long long)lkb->time_bast, lkb->ownpid);
 
 	return buf;
 }
@@ -671,9 +782,8 @@ static void print_lkb(char *line, struct rinfo *ri)
 {
 	struct lkb lkb;
 	char type[4];
-	int rv;
 
-	rv = sscanf(line, "%s %x %d %x %u %llu %x %x %d %d %d %d %d %d %u %llu %llu",
+	sscanf(line, "%s %x %d %x %u %llu %x %x %d %d %d %d %d %d %u %llu %llu",
 		    type,
 		    &lkb.id,
 		    &lkb.nodeid,
@@ -1375,6 +1485,8 @@ static void do_dump(int op)
 		dlmc_dump_debug(buf);
 	else if (op == OP_DUMP_CONFIG)
 		dlmc_dump_config(buf);
+	else if (op == OP_DUMP_RUN)
+		dlmc_dump_run(buf);
 
 	buf[DLMC_DUMP_SIZE-1] = '\0';
 
@@ -1382,6 +1494,22 @@ static void do_dump(int op)
 	printf("\n");
 }
 
+static void do_reload_config(void)
+{
+	if (dlmc_reload_config() < 0)
+		printf("reload_config failed\n");
+	else
+		printf("reload_config done\n");
+}
+
+static void do_set_config(void)
+{
+	if (dlmc_set_config(run_command) < 0)
+		printf("set_config failed\n");
+	else
+		printf("set_config done\n");
+}
+
 static void do_log_plock(void)
 {
 	char buf[DLMC_DUMP_SIZE];
@@ -1396,6 +1524,50 @@ static void do_log_plock(void)
 	printf("\n");
 }
 
+static int do_run(int op)
+{
+	int do_start = (op == OP_RUN) || (op == OP_RUN_START);
+	int do_check = (op == OP_RUN) || (op == OP_RUN_CHECK);
+	int do_cancel = (op == OP_RUN_CANCEL);
+	uint32_t flags = 0;
+	uint32_t check_status = 0;
+	int rv = 0;
+
+	if (do_start) {
+		/* FIXME: use proper option to specify */
+		flags = DLMC_FLAG_RUN_START_NODE_NONE;
+		if (ls_all_nodes)
+			flags = DLMC_FLAG_RUN_START_NODE_RECV;
+
+		rv = dlmc_run_start(run_command, strlen(run_command), 0,
+				    flags, run_uuid);
+
+		printf("run_uuid: %s\n", run_uuid);
+	}
+
+	if (do_check || do_cancel) {
+		if (do_cancel)
+			flags = DLMC_FLAG_RUN_CHECK_CANCEL;
+		else
+			flags = DLMC_FLAG_RUN_CHECK_CLEAR;
+
+		rv = dlmc_run_check(run_uuid, strlen(run_uuid), wait_sec, flags,
+				    &check_status);
+
+		printf("check_status: ");
+
+		if (check_status & DLMC_RUN_STATUS_WAITING)
+			printf("waiting ");
+		if (check_status & DLMC_RUN_STATUS_DONE)
+			printf("done ");
+		if (check_status & DLMC_RUN_STATUS_FAILED)
+			printf("failed ");
+		printf("\n");
+	}
+
+	return rv;
+}
+
 int main(int argc, char **argv)
 {
 	prog_name = argv[0];
@@ -1436,6 +1608,14 @@ int main(int argc, char **argv)
 		do_dump(operation);
 		break;
 
+	case OP_RELOAD_CONFIG:
+		do_reload_config();
+		break;
+
+	case OP_SET_CONFIG:
+		do_set_config();
+		break;
+
 	case OP_LOG_PLOCK:
 		do_log_plock();
 		break;
@@ -1461,6 +1641,17 @@ int main(int argc, char **argv)
 	case OP_FENCE_ACK:
 		do_fence_ack(lsname);
 		break;
+
+	case OP_RUN:
+	case OP_RUN_START:
+	case OP_RUN_CHECK:
+	case OP_RUN_CANCEL:
+		return do_run(operation);
+		break;
+
+	case OP_DUMP_RUN:
+		do_dump(operation);
+		break;
 	}
 	return 0;
 }
